@@ -6,7 +6,6 @@ import functools
 import math
 import numbers
 import operator
-import os
 from collections import OrderedDict
 try:
     from collections.abc import Iterable
@@ -41,21 +40,23 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
 
     @classmethod
     def offsets2parents(cls, offsets):
-        out = cls.numpy.zeros(offsets[-1], dtype=cls.JaggedArray.fget(None).INDEXTYPE)
-        cls.numpy.add.at(out, offsets[offsets != offsets[-1]][1:], 1)
-        cls.numpy.cumsum(out, out=out)
-        if offsets[0] > 0:
-            out[:offsets[0]] = -1
-        return out
+        dtype = cls.JaggedArray.fget(None).INDEXTYPE
+        counts = cls.numpy.diff(cls.numpy.r_[dtype.type(0), offsets])  # numpy >= 1.16: diff prepend
+        indices = cls.numpy.arange(-1, len(offsets) - 1, dtype=dtype)
+        return cls.numpy.repeat(indices, awkward.util.windows_safe(counts))
 
     @classmethod
     def startsstops2parents(cls, starts, stops):
-        out = cls.numpy.full(stops.max(), -1, dtype=cls.JaggedArray.fget(None).INDEXTYPE)
-        lenstarts = len(starts)
-        i = 0
-        while i < lenstarts:
-            out[starts[i]:stops[i]] = i
-            i += 1
+        dtype = cls.JaggedArray.fget(None).INDEXTYPE
+        out = cls.numpy.full(stops.max(), -1, dtype=dtype)
+        indices = cls.numpy.arange(len(starts), dtype=dtype)
+        counts = stops - starts
+        parents = cls.numpy.repeat(indices, awkward.util.windows_safe(counts))
+        contiguous_offsets = cls.JaggedArray.fget(None).counts2offsets(counts)
+        content_indices = cls.numpy.arange(contiguous_offsets[-1])
+        content_indices -= cls.numpy.repeat(contiguous_offsets[:-1], awkward.util.windows_safe(counts))
+        content_indices += starts[parents]
+        cls.numpy.put(out, awkward.util.windows_safe(content_indices), parents)
         return out
 
     @classmethod
@@ -68,7 +69,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         changes[-1] = len(parents)
         changes[1:-1] = tmp
 
-        length = parents.max() + 1 if parents.size > 0 else 0
+        length = parents.max() + 1 if len(parents) > 0 else 0
         starts = cls.numpy.zeros(length, dtype=cls.JaggedArray.fget(None).INDEXTYPE)
         counts = cls.numpy.zeros(length, dtype=cls.JaggedArray.fget(None).INDEXTYPE)
 
@@ -170,7 +171,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         return out
 
     @classmethod
-    def fromindex(cls, index, content, validate=True):
+    def fromlocalindex(cls, index, content, validate=True):
         index = cls._util_toarray(index, cls.INDEXTYPE, (cls.numpy.ndarray, JaggedArray))
         original_counts = None
         if isinstance(index, JaggedArray):
@@ -179,13 +180,13 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             index = index.flatten()
 
         if not cls._util_isintegertype(index.dtype.type):
-            raise TypeError("index must have integer dtype")
+            raise TypeError("localindex must have integer dtype")
         if len(index.shape) != 1 or len(index) != len(content):
-            raise ValueError("index array must be one-dimensional with the same length as content")
+            raise ValueError("localindex array must be one-dimensional with the same length as content")
 
         if validate:
             if not ((index[1:] - index[:-1])[(index != 0)[1:]] == 1).all():
-                raise ValueError("every index that is not zero must be one greater than the previous")
+                raise ValueError("every localindex that is not zero must be one greater than the previous")
 
         starts = cls.numpy.nonzero(index == 0)[0]
         offsets = cls.numpy.empty(len(starts) + 1, dtype=cls.INDEXTYPE)
@@ -193,7 +194,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         offsets[-1] = len(index)
         if original_counts is not None:
             if not cls.numpy.array_equal(offsets[1:] - starts, original_counts):
-                raise ValueError("jagged structure of index does not match jagged structure derived from index")
+                raise ValueError("jagged structure of index does not match jagged structure derived from localindex")
 
         return cls.fromoffsets(offsets, content)
 
@@ -266,19 +267,20 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         else:
             return self.copy(content=self._content.ones_like(**overrides))
 
-    def __awkward_persist__(self, ident, fill, prefix, suffix, schemasuffix, storage, compression, **kwargs):
-        self._valid()
-        if self.offsetsaliased(self._starts, self._stops) and len(self._starts) > 0 and self._starts[0] == 0:
-            return {"id": ident,
-                    "call": ["awkward", "JaggedArray", "fromcounts"],
-                    "args": [fill(self.counts, "JaggedArray.counts", prefix, suffix, schemasuffix, storage, compression, **kwargs),
-                             fill(self._content, "JaggedArray.content", prefix, suffix, schemasuffix, storage, compression, **kwargs)]}
+    def __awkward_serialize__(self, serializer):
+        if self._canuseoffset():
+            return serializer.encode_call(
+                ["awkward", "JaggedArray", "fromcounts"],
+                serializer(self.counts, "JaggedArray.counts"),
+                serializer(self._content, "JaggedArray.content"),
+            )
         else:
-            return {"id": ident,
-                    "call": ["awkward", "JaggedArray"],
-                    "args": [fill(self._starts, "JaggedArray.starts", prefix, suffix, schemasuffix, storage, compression, **kwargs),
-                             fill(self._stops, "JaggedArray.stops", prefix, suffix, schemasuffix, storage, compression, **kwargs),
-                             fill(self._content, "JaggedArray.content", prefix, suffix, schemasuffix, storage, compression, **kwargs)]}
+            return serializer.encode_call(
+                ["awkward", "JaggedArray"],
+                serializer(self._starts, "JaggedArray.starts"),
+                serializer(self._stops, "JaggedArray.stops"),
+                serializer(self._content, "JaggedArray.content"),
+            )
 
     @property
     def starts(self):
@@ -407,10 +409,18 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         self._parents = value
 
     @property
-    def index(self):
-        tmp = self.compact()
-        out = self.numpy.arange(len(tmp._content), dtype=self.INDEXTYPE)
-        return self.JaggedArray(tmp._starts, tmp._stops, (out - tmp._starts[tmp.parents]))
+    def localindex(self):
+        if len(self) == 0:
+            return self.JaggedArray([], [], self.numpy.array([], dtype=self.INDEXTYPE))
+        elif self._canuseoffset():
+            out = self.numpy.arange(self.offsets[0], self.offsets[-1], dtype=self.INDEXTYPE)
+            out -= self.offsets[self.parents[self.parents >= 0]]
+            return self.JaggedArray.fromoffsets(self.offsets - self.offsets[0], out)
+        else:
+            offsets = self.counts2offsets(self.counts)
+            out = self.numpy.arange(offsets[-1], dtype=self.INDEXTYPE)
+            out -= self.numpy.repeat(offsets[:-1], awkward.util.windows_safe(self.counts))
+            return self.JaggedArray.fromoffsets(offsets, out)
 
     def _getnbytes(self, seen):
         if id(self) in seen:
@@ -428,6 +438,14 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
     def _gettype(self, seen):
         return awkward.type.ArrayType(*(self._starts.shape[1:] + (self.numpy.inf, awkward.type._fromarray(self._content, seen))))
 
+    def _util_layout(self, position, seen, lookup):
+        awkward.type.LayoutNode(self._starts, position + (0,), seen, lookup)
+        awkward.type.LayoutNode(self._stops, position + (1,), seen, lookup)
+        awkward.type.LayoutNode(self._content, position + (2,), seen, lookup)
+        return (awkward.type.LayoutArg("starts", position + (0,)),
+                awkward.type.LayoutArg("stops", position + (1,)),
+                awkward.type.LayoutArg("content", position + (2,)))
+
     def _valid(self):
         if not self._isvalid:
             if self.offsetsaliased(self._starts, self._stops):
@@ -538,7 +556,14 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
                 offsets = self.counts2offsets(intheadsum)
 
                 headcontent = self.numpy.array(head._content, dtype=self.BOOLTYPE)
-                headcontent[head.parents < 0] = False
+
+                headcontent_indices_to_ignore = self.numpy.resize(head.parents < 0, headcontent.shape)
+                headcontent_indices_to_ignore[len(head.parents):] = True
+                headcontent[headcontent_indices_to_ignore] = False
+
+                original_headcontent_length = len(headcontent)
+                headcontent = self.numpy.resize(headcontent, thyself._content.shape)
+                headcontent[original_headcontent_length:] = False
 
                 return self.copy(starts=offsets[:-1].reshape(intheadsum.shape), stops=offsets[1:].reshape(intheadsum.shape), content=thyself._content[headcontent])
 
@@ -736,6 +761,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             self._content[where] = self.tojagged(what)._content
 
         elif self._util_isstringslice(where):
+            what = what.unzip()
             if len(where) != len(what):
                 raise ValueError("number of keys ({0}) does not match number of provided arrays ({1})".format(len(where), len(what)))
             for x, y in zip(where, what):
@@ -754,8 +780,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
                 return self.copy(content=data._content)
 
             tmp = self.compact()
-            assert self.offsetsaliased(tmp._starts, tmp._stops)   # because that's what compact means
-            tmpparents = self.offsets2parents(tmp._starts.base)
+            tmpparents = self.offsets2parents(tmp.offsets)
 
             index = self.JaggedArray(tmp._starts, tmp._stops, (self.numpy.arange(tmp._stops[-1], dtype=self.INDEXTYPE) - tmp._starts[tmpparents]))
 
@@ -785,6 +810,9 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         elif isinstance(data, Iterable):
             return self.tojagged(self.numpy.array(data))
 
+        else:
+            return self.tojagged(self.numpy.array([data]))
+
     def _tojagged(self, starts=None, stops=None, copy=True):
         if starts is None and stops is None:
             if copy:
@@ -968,7 +996,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             return out.reshape(self._starts.shape + (count,) + content.shape[1:])
 
         else:
-            indexes = self.numpy.repeat(self._starts, count).reshape(self._starts.shape + (count,))
+            indexes = self.numpy.repeat(self._starts, awkward.util.windows_safe(count)).reshape(self._starts.shape + (count,))
             indexes += self.numpy.arange(count)
             return content[indexes]
 
@@ -1065,14 +1093,14 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         elif n <= 1:
             raise ValueError("Choosing 0 or 1 items is trivial")
 
-        local = np.arange(self.offsets[-1]) - np.repeat(self.offsets[:-1], self.counts)
+        local = self.localindex.content
         offsets = self.JaggedArray.counts2offsets(counts)
         indices = np.arange(offsets[-1])
         parents = self.JaggedArray.offsets2parents(offsets)
 
         if n == 5:
             k5 = indices - offsets[parents]
-            i5 = np.repeat(local, local*(local - 1)*(local - 2)*(local - 3)//24)
+            i5 = np.repeat(local, awkward.util.windows_safe(local*(local - 1)*(local - 2)*(local - 3)//24))
             k4 = k5 - i5*(i5 - 1)*(i5 - 2)*(i5 - 3)*(i5 - 4)//120
             i4 = root4(k4)
             k3 = k4 - i4*(i4 - 1)*(i4 - 2)*(i4 - 3)//24
@@ -1088,7 +1116,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             out = self.JaggedArray.fromoffsets(offsets, self.Table.named("tuple", i1, i2, i3, i4, i5))
         elif n == 4:
             k4 = indices - offsets[parents]
-            i4 = np.repeat(local, local*(local - 1)*(local - 2)//6)
+            i4 = np.repeat(local, awkward.util.windows_safe(local*(local - 1)*(local - 2)//6))
             k3 = k4 - i4*(i4 - 1)*(i4 - 2)*(i4 - 3)//24
             i3 = root3(k3)
             k2 = k3 - i3*(i3 - 1)*(i3 - 2)//6
@@ -1102,7 +1130,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             out = self.JaggedArray.fromoffsets(offsets, self.Table.named("tuple", i1, i2, i3, i4))
         elif n == 3:
             k3 = indices - offsets[parents]
-            i3 = np.repeat(local, local*(local - 1)//2)
+            i3 = np.repeat(local, awkward.util.windows_safe(local*(local - 1)//2))
             k2 = k3 - i3*(i3 - 1)*(i3 - 2)//6
             i2 = root2(k2)
             k1 = k2 - i2*(i2 - 1)//2
@@ -1114,7 +1142,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             out = self.JaggedArray.fromoffsets(offsets, self.Table.named("tuple", i1, i2, i3))
         elif n == 2:
             k2 = indices - offsets[parents]
-            i2 = np.repeat(local, local)
+            i2 = np.repeat(local, awkward.util.windows_safe(local))
             k1 = k2 - i2*(i2 - 1)//2
             i1 = k1
             if absolute:
@@ -1157,7 +1185,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         out = self._argdistincts(absolute=False)
 
         if nested:
-            out = self.JaggedArray.fromcounts(self.numpy.maximum(0, self.counts - 1), self.JaggedArray.fromcounts(self.index[:, :0:-1].flatten(), out._content))
+            out = self.JaggedArray.fromcounts(self.numpy.maximum(0, self.counts - 1), self.JaggedArray.fromcounts(self.localindex[:, :0:-1].flatten(), out._content))
 
         return out
 
@@ -1177,7 +1205,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         out._parents = argpairs._parents
 
         if nested:
-            out = self.JaggedArray.fromcounts(self.numpy.maximum(0, self.counts - 1), self.JaggedArray.fromcounts(self.index[:, :0:-1].flatten(), out._content))
+            out = self.JaggedArray.fromcounts(self.numpy.maximum(0, self.counts - 1), self.JaggedArray.fromcounts(self.localindex[:, :0:-1].flatten(), out._content))
 
         return out
 
@@ -1187,7 +1215,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         out["1"] = out["1"] - self._starts
 
         if nested:
-            out = self.JaggedArray.fromcounts(self.counts, self.JaggedArray.fromcounts((self.index[:, ::-1] + 1).flatten(), out._content))
+            out = self.JaggedArray.fromcounts(self.counts, self.JaggedArray.fromcounts((self.localindex[:, ::-1] + 1).flatten(), out._content))
 
         return out
 
@@ -1200,7 +1228,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         out._parents = argpairs._parents
 
         if nested:
-            out = self.JaggedArray.fromcounts(self.counts, self.JaggedArray.fromcounts((self.index[:, ::-1] + 1).flatten(), out._content))
+            out = self.JaggedArray.fromcounts(self.counts, self.JaggedArray.fromcounts((self.localindex[:, ::-1] + 1).flatten(), out._content))
 
         return out
 
@@ -1297,25 +1325,36 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             out.stops.shape = self._starts.shape  # intentional: self._stops can too long
             return out
 
+    def flattentuple(self):
+        return self.copy(content=self._util_flattentuple(self._content))
+
     def flatten(self, axis=0):
         if not self._util_isinteger(axis) or axis < 0:
             raise TypeError("axis must be a non-negative integer (can't count from the end)")
-        if axis > 0:
-            if isinstance(self._content, JaggedArray):
-                counts = self.JaggedArray.fromcounts(self.counts, self._content.counts).sum()
-                return self.JaggedArray.fromcounts(counts, self._content.flatten(axis=axis - 1))
 
-        if len(self) == 0:
-            return self._content[0:0]
-        elif self._canuseoffset():
-            return self._content[self._starts[0]:self._stops[-1]]
+        if axis > 1:
+            return self.copy(content=self._util_flatten(self._content, axis - 1))
+
+        elif axis == 1:
+            self = self.compact()
+            innercounts = self._util_counts(self._content)
+            if (innercounts < 0).any():
+                raise ValueError("cannot flatten an array containing scalar values")
+            counts = self.JaggedArray.fromcounts(self.counts, innercounts).sum()
+            return type(self).fromcounts(counts, self._util_flatten(self._content, axis - 1))
+
         else:
-            offsets = self.counts2offsets(self.counts.reshape(-1))
-            if len(self._starts.shape) == 1:
-                out = self
+            if len(self) == 0:
+                return self._content[0:0]
+            elif self._canuseoffset():
+                return self._content[self._starts[0]:self._stops[-1]]
             else:
-                out = self.JaggedArray(self._starts.reshape(-1), self._stops.reshape(-1), self._content)
-            return out._tojagged(offsets[:-1], offsets[1:], copy=False)._content
+                offsets = self.counts2offsets(self.counts.reshape(-1))
+                if len(self._starts.shape) == 1:
+                    out = self
+                else:
+                    out = self.JaggedArray(self._starts.reshape(-1), self._stops.reshape(-1), self._content)
+                return out._tojagged(offsets[:-1], offsets[1:], copy=False)._content
 
     def structure1d(self, levellimit=None):
         if len(self._starts) > len(self._stops):
@@ -1345,26 +1384,26 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
     def _hasjagged(self):
         return True
 
-    def _reduce(self, ufunc, identity, dtype, regularaxis):
+    def _reduce(self, ufunc, identity, dtype):
         import awkward.array.table
         self._valid()
 
         if self._util_hasjagged(self._content):
-            return self.copy(content=self._content._reduce(ufunc, identity, dtype, regularaxis))
+            return self.copy(content=self._content._reduce(ufunc, identity, dtype))
 
         elif isinstance(self._content, awkward.array.table.Table):
             out = self._content.copy(contents=[])
             for n, x in self._content._contents.items():
-                out[n] = self.copy(content=x)._reduce(ufunc, identity, dtype, regularaxis)
+                out[n] = self.copy(content=x)._reduce(ufunc, identity, dtype)
             return out
 
         elif isinstance(self._content, awkward.array.base.AwkwardArray):
-            thyself = self.copy(content=self._content._prepare(identity, dtype))
+            thyself = self.copy(content=self._content._prepare(ufunc, identity, dtype))
 
-        elif regularaxis is not None and regularaxis != 0:
+        elif len(self._content.shape) > 1:
             if ufunc is None:
                 ufunc = self.numpy.add
-                if isinstance(self._content, (self.numpy.floating, self.numpy.complexfloating)):
+                if issubclass(self._content.dtype.type, (self.numpy.floating, self.numpy.complexfloating)):
                     content = 1 - self.numpy.isnan(self._content).astype(self.INDEXTYPE)
                 else:
                     content = self.numpy.ones(self._content.shape, dtype=self.INDEXTYPE)
@@ -1379,11 +1418,17 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             if dtype is not None:
                 content = content.astype(dtype)
 
-            return self.copy(content=ufunc.reduce(content, axis=regularaxis))
+            return self.copy(content=ufunc.reduce(content, axis=-1))
 
         else:
             thyself = self.copy()
 
+        if isinstance(thyself._content, awkward.array.table.Table):
+            out = thyself._content.copy(contents=[])
+            for n, x in thyself._content._contents.items():
+                out[n] = thyself.copy(content=x)._reduce(ufunc, identity, dtype)
+            return out
+
         if len(thyself._starts.shape) > 1:
             thyself._starts = thyself._starts.reshape(-1)
             thyself._stops = thyself._stops.reshape(-1)
@@ -1395,7 +1440,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         content = thyself._content
         if ufunc is None:
             ufunc = self.numpy.add
-            if isinstance(content, (self.numpy.floating, self.numpy.complexfloating)):
+            if issubclass(content.dtype.type, (self.numpy.floating, self.numpy.complexfloating)):
                 content = 1 - self.numpy.isnan(content).astype(self.INDEXTYPE)
             else:
                 content = self.numpy.ones(content.shape, dtype=self.INDEXTYPE)
@@ -1404,7 +1449,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             ufunc = self.numpy.add
             content = 1 - (content == 0).astype(self.INDEXTYPE)
 
-        elif isinstance(content, (self.numpy.floating, self.numpy.complexfloating)):
+        elif issubclass(content.dtype.type, (self.numpy.floating, self.numpy.complexfloating)):
             mask = self.numpy.isnan(content)
             if mask.any():
                 content = content.copy()
@@ -1429,21 +1474,15 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
             elif self._util_isintegertype(dtype.type):
                 identity = self.numpy.iinfo(dtype.type).min
 
-        if regularaxis is None:
-            out = self.numpy.empty(thyself._starts.shape[:1], dtype=dtype)
-        else:
-            out = self.numpy.empty(thyself._starts.shape[:1] + content.shape[1:], dtype=dtype)
+        out = self.numpy.empty(thyself._starts.shape[:1], dtype=dtype)
 
         if len(out) != 0:
             nonterminal = thyself.offsets[thyself.offsets < len(content)]
-            if os.name == "nt":    # Windows Numpy reduceat requires 32-bit indexes
-                nonterminal = nonterminal.astype(self.numpy.int32)
 
-            if regularaxis is None:
-                for axis in range(1, len(content.shape)):
-                    content = ufunc.reduce(content, axis=axis)
+            for axis in range(1, len(content.shape)):
+                content = ufunc.reduce(content, axis=axis)
 
-            out = ufunc.reduceat(content, nonterminal)[:len(out)]
+            out = ufunc.reduceat(content, awkward.util.windows_safe(nonterminal))[:len(out)]
             if len(out) < len(thyself):
                 tmp = self.numpy.empty((len(thyself),) + out.shape[1:], dtype=out.dtype)
                 tmp[:len(out)] = out
@@ -1451,10 +1490,7 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
 
             out[thyself.starts == thyself.stops] = identity
 
-        if regularaxis is None:
-            return out.reshape(self._starts.shape)
-        else:
-            return out.reshape(self._starts.shape + self._content.shape[1:])
+        return out.reshape(self._starts.shape)
 
     def argmin(self):
         self._valid()
@@ -1477,6 +1513,9 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         if len(self._content.shape) != 1:
             raise ValueError("cannot compute arg{0} because content is not one-dimensional".format("min" if ismin else "max"))
 
+        if len(self._content) == 0:
+            return self.copy(content=self.numpy.array([], dtype=self.INDEXTYPE))
+
         contentmax = self._content.max()
         shiftval = self.numpy.ceil(contentmax) + 1
         if math.isnan(shiftval) or math.isinf(shiftval) or shiftval != contentmax:
@@ -1587,7 +1626,11 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         # find most general type with a tentative sum which implements the right type-promotion,
         # except for booleans which would get promoted to integers when summing
         def get_dtype(arrays):
-            dtype = np.dtype(sum([x[0] for x in arrays if len(x) != 0]), False)
+            ick = [x[0] for x in arrays if len(x) != 0]
+            if len(ick) == 0:
+                dtype = np.dtype(arrays[0].dtype)
+            else:
+                dtype = np.dtype(sum(ick), False)
             allbools = not np.any([a.dtype != np.dtype(bool) for a in arrays])
             dtype = np.dtype(bool) if allbools else dtype
             return dtype
@@ -1777,13 +1820,21 @@ class JaggedArray(awkward.array.base.AwkwardArrayWithContent):
         else:
             return self.copy(starts=starts, stops=stops, content=self.MaskedArray(mask, content, maskedwhen=maskedwhen))
 
-    def _util_pandas(self, seen):
+    def boolmask(self, maskedwhen=True):
+        if maskedwhen:
+            return self.numpy.zeros(len(self), dtype=self.MASKTYPE)
+        else:
+            return self.numpy.ones(len(self), dtype=self.MASKTYPE)
+
+    _topandas_name = "JaggedSeries"
+
+    def _topandas(self, seen):
         import awkward.pandas
         if id(self) in seen:
             return seen[id(self)]
         else:
             out = seen[id(self)] = self.copy()
-            out.__class__ = awkward.pandas.mixin("JaggedSeries", self)
+            out.__class__ = awkward.pandas.mixin(type(self))
             if isinstance(self._content, awkward.array.base.AwkwardArray):
-                out._content = out._content._util_pandas(seen)
+                out._content = out._content._topandas(seen)
             return out
